package com.softslate.commerce.businessobjects.administrator;

import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.softslate.commerce.businessobjects.core.BaseBusinessProcessor;
import com.softslate.commerce.daos.administrator.AdministratorDAO;
import com.softslate.commerce.daos.administrator.AdministratorGatewayDAO;

/**
 * Processes business logic having to do with administrators. Created in the
 * Struts layer in response to requests for viewing, editing, adding, and
 * deleting administrators.
 * <p>
 * <code>BasicAdministratorProcessor</code> is the default implementation of the
 * <code>AdministratorProcessor</code> interface for the application.
 * 
 * @author David Tobey
 */
public class BasicAdministratorProcessor extends BaseBusinessProcessor implements AdministratorProcessor {

	static Log log = LogFactory.getLog(BasicAdministratorProcessor.class);

	public Map processLogin(Map parameters) throws Exception {
		Map result = new HashMap();
		if (getUser().getAdministrator() != null) {
			if (log.isInfoEnabled())
				log.info("Administrator is already logged in.");
			result.put("resultCode", "errors.userAlreadyLoggedIn");
			return result;
		}
		AdministratorDAO administratorDAO =  getDaoFactory().createDAO(AdministratorDAO.class);
		AdministratorGatewayDAO administratorGatewayDAO = (AdministratorGatewayDAO) getDaoFactory().createDAO(
				"administratorGatewayDAOImplementer");
		int numberOfPreviousPasswordsToCheck = 0;
		String numberOfPreviousPasswordsToCheckStr = getSettings().getValue("numberOfPreviousPasswordsToCheck");
		if (numberOfPreviousPasswordsToCheckStr != null) {
			numberOfPreviousPasswordsToCheck = Integer.parseInt(getSettings().getValue(
					"numberOfPreviousPasswordsToCheck"));
		}
		String userName = (String) parameters.get("userName");
		String password = (String) parameters.get("password");
		String newDecryptedPassword = (String) parameters.get("newDecryptedPassword");
		String ipAddress = (String) parameters.get("ipAddress");
		String newPasswordEncrypted = null;
		if (newDecryptedPassword != null) {
			newPasswordEncrypted = administratorDAO.getDAOUtils().getEncrypted(newDecryptedPassword,
					(String) getSettings().getValue("administratorEncryptionType"),
					getAppSettings().getProperty("twoWayKeyFile"),
					(String) getSettings().getValue("oneWayEncryptionAlgorithm"));
		}
		result = administratorGatewayDAO.processLogin(userName, password);
		Administrator validAdministrator = (Administrator) result.get("validAdministrator");
		if (validAdministrator == null) {
			getUser().setAdministrator(null);
			Administrator administrator = administratorDAO.loadAdministratorByUserName(userName);
			if (administrator != null) {
				resetNumberOfFailedLoginsIfLockOutExpired(administrator);
				incrementNumberOfFailedLogins(administrator);
			}
			if (administrator != null && !isAdministratorLockedOut(administrator)) {
				lockOutIfTooManyFailedLogins(administrator);
			}

			// Whether the login was valid or not, show the locked out error if the user is locked out.
			if (isAdministratorLockedOut(administrator)) {
				result.put("resultCode", "errors.administratorLockedOut");
			}
		} else if (isAdministratorLockedOut(validAdministrator)) {
			result.put("resultCode", "errors.administratorLockedOut");
			result.put("validAdministrator", null);
			getUser().setAdministrator(null);
		} else if (!isValidIPAddress(validAdministrator, ipAddress)) {
			result.put("resultCode", "errors.invalidIPAddress");
			result.put("validAdministrator", null);
			getUser().setAdministrator(null);
		} else if (newDecryptedPassword == null && isPasswordExpired(validAdministrator)) {
			result.put("resultCode", "errors.passwordHasExpired");
			result.put("validAdministrator", null);
			getUser().setAdministrator(null);
		} else if (newDecryptedPassword != null
				&& isSameAsPreviousPassword(validAdministrator, newPasswordEncrypted, numberOfPreviousPasswordsToCheck)) {
			result.put("resultCode", "errors.sameAsPreviousPassword");
			result.put("numberOfPreviousPasswordsToCheck", numberOfPreviousPasswordsToCheck);
			result.put("validAdministrator", null);
			getUser().setAdministrator(null);
		} else {
			validAdministrator.setLastLogin(formatDateTime(new Date()));
			validAdministrator.setNumberOfFailedLogins(0);
			validAdministrator.setLockedOutUntil(null);
			if (newDecryptedPassword != null) {
				administratorChangePassword(newDecryptedPassword, validAdministrator);
			}
			getUser().setAdministrator(validAdministrator);
		}
		return result;
	}

	protected boolean isAdministratorLockedOut(Administrator administrator) {
		boolean ret = false;
		if (administrator != null && administrator.getLockedOutUntil() != null
				&& !administrator.getLockedOutUntil().equals("")) {
			Date lockedOutUntilDate = parseDateTime(administrator.getLockedOutUntil());
			Date now = new Date();
			if (now.before(lockedOutUntilDate)) {
				log.info("Administrator " + administrator.getUserName() + " is locked out until: "
						+ administrator.getLockedOutUntil());
				ret = true;
			}
		}
		return ret;
	}

	protected void resetNumberOfFailedLoginsIfLockOutExpired(Administrator administrator) {
		if (administrator != null && administrator.getLockedOutUntil() != null
				&& !administrator.getLockedOutUntil().equals("")) {
			Date lockedOutUntilDate = parseDateTime(administrator.getLockedOutUntil());
			Date now = new Date();
			if (now.after(lockedOutUntilDate)) {
				log.info("Administrator " + administrator.getUserName()
						+ "'s lock out period expired - resetting number of failed logins.");
				administrator.setNumberOfFailedLogins(0);
				administrator.setLockedOutUntil(null);
			}
		}
	}

	protected void incrementNumberOfFailedLogins(Administrator administrator) {
		int numberOfFailedLogins = 0;
		if (administrator.getNumberOfFailedLogins() != null) {
			numberOfFailedLogins = administrator.getNumberOfFailedLogins();
		}
		administrator.setNumberOfFailedLogins(++numberOfFailedLogins);
	}

	protected void lockOutIfTooManyFailedLogins(Administrator administrator) {
		int numFailedLoginsBeforeLockingOut = 0;
		String numberFailedLoginsBeforeLockingOut = getSettings().getValue("numberFailedLoginsBeforeLockingOut");
		if (numberFailedLoginsBeforeLockingOut != null && administrator.getNumberOfFailedLogins() != null) {
			numFailedLoginsBeforeLockingOut = Integer.parseInt(numberFailedLoginsBeforeLockingOut);
		}
		if (numFailedLoginsBeforeLockingOut > 0
				&& administrator.getNumberOfFailedLogins() >= numFailedLoginsBeforeLockingOut) {
			String lockoutDurationInMinutes = getSettings().getValue("lockoutDurationInMinutes");
			Calendar cal = Calendar.getInstance();
			cal.add(Calendar.MINUTE, Integer.parseInt(lockoutDurationInMinutes));
			String lockedOutUntil = this.formatDateTime(new Date(cal.getTimeInMillis()));
			administrator.setLockedOutUntil(lockedOutUntil);
		}
	}

	protected boolean isPasswordExpired(Administrator administrator) {
		boolean ret = false;
		if (administrator.getPasswordExpires() != null && administrator.getPasswordExpires().length() > 0) {
			Date passwordExpiredDate = parseDateTime(administrator.getPasswordExpires());
			Date now = new Date();
			if (now.after(passwordExpiredDate)) {
				log.info("Administrator password has expired: " + administrator.getUserName() + " (expired "
						+ administrator.getPasswordExpires() + ").");
				ret = true;
			}
		}
		return ret;
	}

	protected boolean isValidIPAddress(Administrator administrator, String ipAddress) {
		boolean ret = true;
		if (administrator.getAllowedIPAddresses() != null && administrator.getAllowedIPAddresses().trim().length() > 0) {
			ret = false;
			String[] ipAddresses = administrator.getAllowedIPAddresses().split(",");
			for (int i = 0; i < ipAddresses.length; i++) {
				String ipAddr = ipAddresses[i].trim();
				if (ipAddr.equals(ipAddress)) {
					ret = true;
				}
			}
		}
		return ret;
	}

	protected boolean isSameAsPreviousPassword(Administrator administrator, String newPasswordEncrypted,
			int numberOfPreviousPasswordsToCheck) {
		boolean ret = false;
		if (administrator.getPreviousPasswords() != null && numberOfPreviousPasswordsToCheck > 0) {
			String[] previousPasswords = administrator.getPreviousPasswords().split(",");
			for (int i = 0; i < previousPasswords.length && i < numberOfPreviousPasswordsToCheck; i++) {
				String previousPassword = previousPasswords[i].trim();
				if (previousPassword.equals(newPasswordEncrypted)) {
					ret = true;
				}
			}
		}
		return ret;
	}

	public Map processApiLogin(String apiLoginID, String apiAuthenticationToken) throws Exception {
		Map result = new HashMap();
		AdministratorGatewayDAO administratorGatewayDAO = (AdministratorGatewayDAO) getDaoFactory().createDAO(
				"administratorGatewayDAOImplementer");
		Administrator administrator = administratorGatewayDAO.processApiLogin(apiLoginID, apiAuthenticationToken);
		getUser().setAdministrator(administrator);
		if (administrator == null) {
			log.warn("Invalid API login - authentication failed for user: " + apiLoginID);
			result.put("resultCode", "errors.invalidLogin");
		} else {
			log.info("Valid API login for user: " + apiLoginID);
			result.put("resultCode", "0");
		}
		return result;
	}

	public Map loadAdministratorsAndCount(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName() + ": Loading administrators and count");
		AdministratorGatewayDAO gatewayDAO = (AdministratorGatewayDAO) getDaoFactory().createDAO(
				"administratorGatewayDAOImplementer");
		return gatewayDAO.loadAdministratorsAndCount(parameters);
	}

	public Map addAdministrator(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			String code = (String) parameters.get("userName");
			log.info(getUser().getAdministrator().getUserName() + ": Adding administrator: " + code);
		}
		Map results = new HashMap();
		Administrator administrator = (Administrator) getBusinessObjectFactory().createObject(
				"administratorImplementer");
		BeanUtils.copyProperties(administrator, parameters);
		AdministratorDAO dao =  getDaoFactory().createDAO(AdministratorDAO.class);
		dao.setAdministrator(administrator);
		dao.insertAdministrator(true);
		results.put("newRecord", dao.getAdministrator());
		return results;
	}

	public Administrator loadAdministratorFromID(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName() + ": Loading administrator from id");
		Administrator administrator = (Administrator) getBusinessObjectFactory().createObject(
				"administratorImplementer");
		AdministratorDAO dao =  getDaoFactory().createDAO(AdministratorDAO.class);
		BeanUtils.copyProperties(administrator, parameters);
		dao.setAdministrator(administrator);
		dao.loadAdministratorFromID();
		return dao.getAdministrator();
	}

	public Map editAdministrator(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			String code = (String) parameters.get("userName");
			log.info(getUser().getAdministrator().getUserName() + ": Editing administrator: " + code);
		}
		Map results = new HashMap();
		Administrator administrator = null;
		administrator = loadAdministratorFromID(parameters);
		BeanUtils.copyProperties(administrator, parameters);
		AdministratorDAO dao =  getDaoFactory().createDAO(AdministratorDAO.class);
		dao.setAdministrator(administrator);
		dao.updateAdministrator(true);
		results.put("updatedRecord", dao.getAdministrator());
		return results;
	}

	public Map editDeleteAdministrators(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName() + ": Editing/deleting multiple administrators");
		AdministratorGatewayDAO dao = (AdministratorGatewayDAO) getDaoFactory().createDAO(
				"administratorGatewayDAOImplementer");
		Map results = dao.editDeleteAdministrators(parameters, getBusinessObjectFactory());
		return results;
	}

	public void deleteAdministrator(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			String code = (String) parameters.get("userName");
			log.info(getUser().getAdministrator().getUserName() + ": Deleting administrator: " + code);
		}
		Administrator administrator = (Administrator) getBusinessObjectFactory().createObject(
				"administratorImplementer");
		BeanUtils.copyProperties(administrator, parameters);
		AdministratorDAO dao =  getDaoFactory().createDAO(AdministratorDAO.class);
		dao.setAdministrator(administrator);
		dao.deleteAdministrator(true);
		return;
	}

	public Map processChangePassword(String currentDecryptedPassword, String newDecryptedPassword, String userName)
			throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			log.info(getUser().getAdministrator().getUserName() + ": Changing administrator password.");
		}
		Map result = new HashMap();
		AdministratorDAO dao =  getDaoFactory().createDAO(AdministratorDAO.class);
		int numberOfPreviousPasswordsToCheck = Integer.parseInt(getSettings().getValue(
				"numberOfPreviousPasswordsToCheck"));
		String newPasswordEncrypted = dao.getDAOUtils().getEncrypted(newDecryptedPassword,
				(String) getSettings().getValue("administratorEncryptionType"),
				getAppSettings().getProperty("twoWayKeyFile"),
				(String) getSettings().getValue("oneWayEncryptionAlgorithm"));
		Administrator administrator = dao.loadFromUserNameAndPassword(userName, currentDecryptedPassword);
		if (administrator == null) {
			result.put("resultCode", "errors.currentPasswordInvalid");
		} else if (newDecryptedPassword != null
				&& isSameAsPreviousPassword(administrator, newPasswordEncrypted, numberOfPreviousPasswordsToCheck)) {
			result.put("resultCode", "errors.sameAsPreviousPassword");
		} else {
			administratorChangePassword(newDecryptedPassword, administrator);
			getUser().getAdministrator().setDecryptedPassword(administrator.getDecryptedPassword());
			getUser().getAdministrator().setPasswordExpires(administrator.getPasswordExpires());
			getUser().getAdministrator().setPreviousPasswords(administrator.getPreviousPasswords());
		}
		return result;
	}

	public void administratorChangePassword(String newDecryptedPassword, Administrator administrator) throws Exception {
		AdministratorDAO dao =  getDaoFactory().createDAO(AdministratorDAO.class);
		String newPasswordEncrypted = dao.getDAOUtils().getEncrypted(newDecryptedPassword,
				(String) getSettings().getValue("administratorEncryptionType"),
				getAppSettings().getProperty("twoWayKeyFile"),
				(String) getSettings().getValue("oneWayEncryptionAlgorithm"));
		String passwordExpires = computeAndFormatPasswordExpiration();
		String previousPasswords = formatPreviousPasswords(administrator, newPasswordEncrypted);
		administrator.setDecryptedPassword(newDecryptedPassword);
		administrator.setPasswordExpires(passwordExpires);
		administrator.setPreviousPasswords(previousPasswords);
		dao.setAdministrator(administrator);
		dao.updateAdministrator(true);
	}

	protected String computeAndFormatPasswordExpiration() {
		String passwordExpires = null;
		int passwordExpirationInDays = Integer.parseInt(getSettings().getValue("passwordExpirationInDays"));
		if (passwordExpirationInDays > 0) {
			Calendar cal = Calendar.getInstance();
			cal.add(Calendar.DATE, passwordExpirationInDays);
			passwordExpires = this.formatDateTime(new Date(cal.getTimeInMillis()));
		}
		return passwordExpires;
	}

	protected String formatPreviousPasswords(Administrator administrator, String newPasswordEncrypted) {
		StringBuffer previousPasswords = new StringBuffer("");
		previousPasswords.append(newPasswordEncrypted);
		previousPasswords.append(",");
		if (administrator.getPreviousPasswords() != null) {
			String[] prevPasswords = administrator.getPreviousPasswords().split(",");
			for (int i = 0; i < prevPasswords.length && i < 20; i++) {
				String previousPassword = prevPasswords[i].trim();
				previousPasswords.append(previousPassword);
				previousPasswords.append(",");
			}
		}
		return previousPasswords.toString();
	}

}
