package com.softslate.commerce.businessobjects.payment;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import urn.ebay.api.PayPalAPI.DoExpressCheckoutPaymentReq;
import urn.ebay.api.PayPalAPI.DoExpressCheckoutPaymentRequestType;
import urn.ebay.api.PayPalAPI.DoExpressCheckoutPaymentResponseType;
import urn.ebay.api.PayPalAPI.GetExpressCheckoutDetailsReq;
import urn.ebay.api.PayPalAPI.GetExpressCheckoutDetailsRequestType;
import urn.ebay.api.PayPalAPI.GetExpressCheckoutDetailsResponseType;
import urn.ebay.api.PayPalAPI.PayPalAPIInterfaceServiceService;
import urn.ebay.api.PayPalAPI.SetExpressCheckoutReq;
import urn.ebay.api.PayPalAPI.SetExpressCheckoutRequestType;
import urn.ebay.api.PayPalAPI.SetExpressCheckoutResponseType;
import urn.ebay.apis.CoreComponentTypes.BasicAmountType;
import urn.ebay.apis.eBLBaseComponents.AbstractResponseType;
import urn.ebay.apis.eBLBaseComponents.AckCodeType;
import urn.ebay.apis.eBLBaseComponents.CurrencyCodeType;
import urn.ebay.apis.eBLBaseComponents.DoExpressCheckoutPaymentRequestDetailsType;
import urn.ebay.apis.eBLBaseComponents.DoExpressCheckoutPaymentResponseDetailsType;
import urn.ebay.apis.eBLBaseComponents.ErrorType;
import urn.ebay.apis.eBLBaseComponents.GetExpressCheckoutDetailsResponseDetailsType;
import urn.ebay.apis.eBLBaseComponents.PaymentActionCodeType;
import urn.ebay.apis.eBLBaseComponents.PaymentDetailsItemType;
import urn.ebay.apis.eBLBaseComponents.PaymentDetailsType;
import urn.ebay.apis.eBLBaseComponents.SetExpressCheckoutRequestDetailsType;
import urn.ebay.apis.eBLBaseComponents.SolutionTypeType;

import com.softslate.commerce.businessobjects.order.Order;
import com.softslate.commerce.businessobjects.order.OrderDiscount;
import com.softslate.commerce.businessobjects.order.OrderItem;
import com.softslate.commerce.daos.payment.PaymentDAO;

/**
 * Processes a payment in the store for a given user's cart through PayPal
 * Express Checkout. The three API calls required by PayPal Express integration
 * are handled in the {@link #processSetExpressCheckout},
 * {@link #processGetExpressCheckoutDetails}, and
 * {@link #processDoExpressCheckoutPayment} methods.
 * 
 * @author David Tobey
 */
public class PayPalExpressProcessor extends BasePaymentProcessor implements
		IPayPalExpressPaymentProcessor {

	static Log log = LogFactory.getLog(PayPalExpressProcessor.class);

	public Map createPaypalConfigMap() throws Exception {

		String paypalAPIUsername = (String) getSettings().getValue("paypalAPIUsername");
		String paypalAPIPassword = (String) getSettings().getValue("paypalAPIPassword");
		String paypalEnvironment = (String) getSettings().getValue("paypalEnvironment");
		String paypalSignature = (String) getSettings().getValue("paypalSignature");
		String paypalSubject = (String) getSettings().getValue("paypalSubject");
		String paypalCertificateFile = (String) getSettings().getValue("paypalCertificateFile");
		String paypalPrivateKeyPassword = (String) getSettings().getValue("paypalPrivateKeyPassword");
		String paypalTimeout = (String) getSettings().getValue("paypalTimeout");
		// String paypalDelayTime = (String) getSettings().getValue(
		// "paypalDelayTime");
		String paypalMaximumRetries = (String) getSettings().getValue("paypalMaximumRetries");

		Map<String, String> configMap = new HashMap<String, String>();

		// Account Credential
		configMap.put("acct1.UserName", paypalAPIUsername);
		configMap.put("acct1.Password", paypalAPIPassword);
		if (paypalCertificateFile != null && paypalCertificateFile.length() > 0 && paypalPrivateKeyPassword != null
				&& paypalPrivateKeyPassword.length() > 0) {
			configMap.put("acct1.CertKey", paypalPrivateKeyPassword);
			configMap.put("acct1.CertPath", paypalCertificateFile);
			// configMap.put("acct2.AppId", "APP-80W284485P519543T");
		} else {
			configMap.put("acct1.Signature", paypalSignature);
		}

		// Subject is optional, only required in case of third party permission
		if (paypalSubject != null && paypalSubject.length() > 0) {
			configMap.put("acct1.Subject", paypalSubject);
		}

		// "live" or "sandbox"
		configMap.put("mode", paypalEnvironment);

		// Connection settings
		configMap.put("http.ConnectionTimeOut", paypalTimeout);
		configMap.put("http.Retry", paypalMaximumRetries);

		// configMap.put("http.ReadTimeOut", paypalTimeout);
		// configMap.put("http.MaxConnection", "100");

		return configMap;
	}

	public Map processSetExpressCheckout(Map parameters) throws Exception {
		if (log.isInfoEnabled())
			log.info("Processing SetExpressCheckout.");

		// Create the profile
		PayPalAPIInterfaceServiceService service = new PayPalAPIInterfaceServiceService(createPaypalConfigMap());
		SetExpressCheckoutRequestType pprequest = new SetExpressCheckoutRequestType();
		SetExpressCheckoutRequestDetailsType details = new SetExpressCheckoutRequestDetailsType();

		// Set the order total and item details
		Double subtotal = getUser().getOrder().getSubtotal();
		String sendCartItemsToPayPal = (String) getSettings().getValue("sendCartItemsToPayPal");
		if (sendCartItemsToPayPal == null || sendCartItemsToPayPal.equals("1")) {
			PaymentDetailsType paymentDetails = new PaymentDetailsType();
			PaymentDetailsItemType[] paypalItems = convertCartToPayPalItems(getUser().getOrder());
			if (paypalItems != null) {
				paymentDetails.setPaymentDetailsItem(Arrays.asList(paypalItems));
				details.setPaymentDetails(java.util.Collections.singletonList(paymentDetails));
				subtotal = determinePayPalOrderTotal(paypalItems);
			}
		}
		subtotal = utils().roundForMoney(subtotal);
		String orderTotalString = subtotal.toString();
		BasicAmountType orderTotal = new BasicAmountType();
		orderTotal.setValue(orderTotalString);
		String currency = (String) getSettings().getValue("currencyCode");
		orderTotal.setCurrencyID(CurrencyCodeType.fromValue(currency));
		details.setOrderTotal(orderTotal);
		if (log.isDebugEnabled())
			log.debug("orderTotalString: " + orderTotalString);

		// Set the transaction type
		String paymentType = (String) getSettings().getValue(
				"paypalPaymentType");
		details.setPaymentAction(PaymentActionCodeType.fromValue(paymentType));

		// Specify that the consumer is not required to have a PayPal account
		details.setSolutionType(SolutionTypeType.SOLE);

		// Set the cancel URL
		String customerURL = (String) getSettings().getValue("customerURL");
		String cancelURL = customerURL + "/Cart.do";
		details.setCancelURL(cancelURL);
		if (log.isDebugEnabled())
			log.debug("cancelURL: " + cancelURL);

		// Set the return URL
		String secureURL = (String) getSettings().getValue("customerSecureURL");
		String returnURL = secureURL + "/PayPalExpressReturn.do";
		details.setReturnURL(returnURL);
		//details.setReturnURL(returnURL + "?amount=" + orderTotalString
				//+ "&currency=" + currency);
		if (log.isDebugEnabled())
			log.debug("returnURL: " + returnURL);

		// Make the SetExpressCheckout call
		pprequest.setSetExpressCheckoutRequestDetails(details);
		SetExpressCheckoutReq expressCheckoutReq = new SetExpressCheckoutReq();
		expressCheckoutReq.setSetExpressCheckoutRequest(pprequest);
		SetExpressCheckoutResponseType ppresponse = service.setExpressCheckout(expressCheckoutReq);

		// Save the response in the user object
		Map results = this.prepareResponseFromSetExpressCheckoutResponseType(ppresponse);
		if (null != ppresponse.getAck()
				&& (ppresponse.getAck().equals(AckCodeType.SUCCESS) || ppresponse.getAck().equals(
						AckCodeType.SUCCESSWITHWARNING))) {

			// Save the token
			if (getUser().getCustomMap() == null)
				getUser().setCustomMap(new HashMap());

			getUser().getCustomMap().put("paypalToken", ppresponse.getToken());
		}
		return results;
	}

	protected Map prepareResponseFromSetExpressCheckoutResponseType(
			AbstractResponseType ppresponse) {
		HashMap responseMap = new HashMap();

		if (null != ppresponse.getAck()
				&& (ppresponse.getAck().equals(AckCodeType.SUCCESS) || ppresponse.getAck().equals(
						AckCodeType.SUCCESSWITHWARNING))) {
			responseMap.put("resultCode", "0");
		} else {
			responseMap.put("resultCode", "-1");
			responseMap.put("paypalAPIVersion", ppresponse.getVersion());
			responseMap.put("paypalCorrId", ppresponse.getCorrelationID());

			if (null != ppresponse.getErrors() && ppresponse.getErrors().size() > 0) {
				ArrayList<String> errorMessages = new ArrayList<String>();
				for (ErrorType error: ppresponse.getErrors()) {
					errorMessages.add(error.getErrorCode() + " : " + error.getLongMessage());
				}
				responseMap.put("resultMessages", errorMessages.toArray(new String[errorMessages.size()]));
			}


		}

		return responseMap;
	}

	public Map processGetExpressCheckoutDetails(Map parameters)
			throws Exception {
		if (log.isInfoEnabled())
			log.info("Processing GetExpressCheckoutDetails.");

		// Check that we have a token
		if (getUser().getCustomMap() == null
				|| getUser().getCustomMap().get("paypalToken") == null) {
			return null;
		}

		// Create the profile
		if (log.isDebugEnabled())
			log.debug("Creating the paypal profile");

		PayPalAPIInterfaceServiceService service = new PayPalAPIInterfaceServiceService(createPaypalConfigMap());

		// Send the request
		if (log.isDebugEnabled())
			log.debug("Sending the paypal request");
		GetExpressCheckoutDetailsReq req = new GetExpressCheckoutDetailsReq();
		GetExpressCheckoutDetailsRequestType reqType = new GetExpressCheckoutDetailsRequestType((String) getUser()
				.getCustomMap().get("paypalToken"));
		req.setGetExpressCheckoutDetailsRequest(reqType);
		GetExpressCheckoutDetailsResponseType ppresponse = service.getExpressCheckoutDetails(req);

		// Save the response in the user object
		if (log.isDebugEnabled())
			log.debug("Saving the paypal response");

		Map results = this
				.prepareResponseFromSetExpressCheckoutResponseType(ppresponse);

		if (ppresponse.getAck().equals(AckCodeType.SUCCESS)
				|| ppresponse.getAck().equals(AckCodeType.SUCCESSWITHWARNING)) {

			// Store the PayPal shipping details in the user's order
			if (log.isDebugEnabled())
				log
						.debug("Successful paypal response - store info in user's order");

			GetExpressCheckoutDetailsResponseDetailsType details = ppresponse
					.getGetExpressCheckoutDetailsResponseDetails();
			if (details.getPayerInfo().getAddress().getName() != null
					&& details.getPayerInfo().getAddress().getName().length() > 0) {
				String firstNameParsed = "";
				String lastNameParsed = "";
				String[] nameParts = details.getPayerInfo().getAddress().getName().split(" ");
				if (nameParts.length > 1) {
					firstNameParsed = nameParts[0];
					lastNameParsed = details.getPayerInfo().getAddress().getName()
							.replaceFirst(firstNameParsed + " ", "");
				} else {
					firstNameParsed = "[PayPal]";
					lastNameParsed = details.getPayerInfo().getAddress().getName();
				}
				getUser().getOrder().firstDelivery().setFirstName(firstNameParsed);
				getUser().getOrder().firstDelivery().setLastName(lastNameParsed);
			} else {
				getUser().getOrder().firstDelivery().setFirstName(
						details.getPayerInfo().getPayerName().getFirstName());
				getUser().getOrder().firstDelivery().setLastName(
						details.getPayerInfo().getPayerName().getLastName());
			}
			getUser().getOrder().firstDelivery().setAddress1(
					details.getPayerInfo().getAddress().getStreet1());
			getUser().getOrder().firstDelivery().setAddress2(
					details.getPayerInfo().getAddress().getStreet2());
			getUser().getOrder().firstDelivery().setCity(
					details.getPayerInfo().getAddress().getCityName());
			String stateOrProvince = convertStateOrProvice(details
					.getPayerInfo().getAddress().getStateOrProvince());
			getUser().getOrder().firstDelivery().setState(stateOrProvince);
			getUser().getOrder().firstDelivery().setPostalCode(
					details.getPayerInfo().getAddress().getPostalCode());
			getUser().getOrder().firstDelivery()
					.setCountry(
							details.getPayerInfo().getAddress().getCountry()
									.toString());
			getUser().getOrder().firstDelivery().setPhone1(
					details.getPayerInfo().getAddress().getPhone());

			// Populate the billing details
			getUser().getOrder().setAddress1("[PayPal]");
			getUser().getOrder().setAddress2("[PayPal]");
			getUser().getOrder().setCity("[PayPal]");
			getUser().getOrder().setState("[PayPal]");
			getUser().getOrder().setPostalCode("[PayPal]");
			getUser().getOrder().setFirstName(
					details.getPayerInfo().getPayerName().getFirstName());
			getUser().getOrder().setLastName(
					details.getPayerInfo().getPayerName().getLastName());
			getUser().getOrder().setOrganization(
					details.getPayerInfo().getPayerBusiness());
			getUser().getOrder().setCountry(
					details.getPayerInfo().getPayerCountry().toString());
			getUser().getOrder().setPhone1(details.getContactPhone());
			getUser().getOrder().setEmail1(details.getPayerInfo().getPayer());

			// Store other custom things
			getUser().getCustomMap().put("paypalPayerID",
					details.getPayerInfo().getPayerID());
			getUser().getCustomMap().put(
					"paypalPayerStatus",
					details.getPayerInfo().getPayerStatus().getValue()
							.toString());
			getUser().getCustomMap().put(
					"paypalAddressStatus",
					details.getPayerInfo().getAddress().getAddressStatus()
							.getValue().toString());

		}

		return results;
	}

	public Map processDoExpressCheckoutPayment(Map parameters) throws Exception {
		if (log.isInfoEnabled())
			log.info("Processing DoExpressCheckoutPayment.");

		// Check that we have a token and a Payer ID
		if (getUser().getCustomMap() == null
				|| getUser().getCustomMap().get("paypalToken") == null
				|| getUser().getCustomMap().get("paypalPayerID") == null) {
			if (log.isInfoEnabled())
				log.info("No token or payer ID for user.");
			return null;
		}

		// Create the profile
		PayPalAPIInterfaceServiceService service = new PayPalAPIInterfaceServiceService(createPaypalConfigMap());

		DoExpressCheckoutPaymentRequestType pprequest = new DoExpressCheckoutPaymentRequestType();
		DoExpressCheckoutPaymentRequestDetailsType details = new DoExpressCheckoutPaymentRequestDetailsType();
		details.setToken((String) getUser().getCustomMap().get("paypalToken"));
		details.setPayerID((String) getUser().getCustomMap().get(
				"paypalPayerID"));

		// Set the transaction type
		String paymentType = (String) getSettings().getValue(
				"paypalPaymentType");

		details.setPaymentAction(PaymentActionCodeType.fromValue(paymentType));

		// Set the currency
		String currency = (String) getSettings().getValue("currencyCode");

		// Set the order total
		String orderTotalString = getUser().getOrder().getTotal().toString();
		PaymentDetailsType paymentDetails = new PaymentDetailsType();
		BasicAmountType orderTotal = new BasicAmountType();
		orderTotal.setValue(orderTotalString);
		orderTotal.setCurrencyID(CurrencyCodeType.fromValue(currency));
		paymentDetails.setOrderTotal(orderTotal);
		paymentDetails.setButtonSource("SoftSlateLLC_Cart");

		// Set the order description
		paymentDetails.setOrderDescription((String) getSettings().getValue(
				"storeName")
				+ "(" + getUser().getOrder().getOrderID() + ")");

		List<PaymentDetailsType> paymentDetailsList = new ArrayList<PaymentDetailsType>();
		paymentDetailsList.add(paymentDetails);
		details.setPaymentDetails(paymentDetailsList);
		pprequest.setDoExpressCheckoutPaymentRequestDetails(details);

		// Send the request
		List<PaymentDetailsType> payDetailType = new ArrayList<PaymentDetailsType>();
		payDetailType.add(paymentDetails);
		details.setPaymentDetails(payDetailType);
		DoExpressCheckoutPaymentReq doExpressCheckoutPaymentReq = new DoExpressCheckoutPaymentReq();
		doExpressCheckoutPaymentReq.setDoExpressCheckoutPaymentRequest(pprequest);
		DoExpressCheckoutPaymentResponseType ppresponse = service.doExpressCheckoutPayment(doExpressCheckoutPaymentReq);
		DoExpressCheckoutPaymentResponseDetailsType responseDetails = ppresponse
				.getDoExpressCheckoutPaymentResponseDetails();

		// Save the response
		Map results = this
				.prepareResponseFromSetExpressCheckoutResponseType(ppresponse);

		if (ppresponse.getAck().equals(AckCodeType.SUCCESS)
				|| ppresponse.getAck().equals(AckCodeType.SUCCESSWITHWARNING)) {

			// Process a successful resposne
			if (parameters == null) {
				parameters = new HashMap();
			}
			parameters.put("orderTotal", orderTotalString);
			parameters.put("transactionID", responseDetails.getPaymentInfo().get(0)
					.getTransactionID());
			parameters.put("token", responseDetails.getToken());
			parameters.put("paymentStatus", responseDetails.getPaymentInfo().get(0)
					.getPaymentStatus().getValue().toString());
			parameters.put("paymentType", responseDetails.getPaymentInfo().get(0)
					.getPaymentType().getValue().toString());
			parameters.put("transactionType", responseDetails.getPaymentInfo().get(0)
					.getTransactionType().getValue().toString());

			if (log.isDebugEnabled())
				log.debug("Result parameters: " + parameters);

			processSuccessfulTransaction(parameters);

		}
		return results;
	}

	public Map process(Map parameters) throws Exception {
		if (log.isInfoEnabled())
			log.info("Processing payment.");

		Map results = processDoExpressCheckoutPayment(parameters);

		return results;
	}

	public void processSuccessfulTransaction(Map parameters) throws Exception {
		Payment payment = (Payment) getBusinessObjectFactory().createObject(
				"paymentImplementer");
		BeanUtils.copyProperties(payment, parameters);
		payment.setStatus("Received");
		payment.setAmount(getUser().getOrder().getTotal());
		payment.setCreated(formatDateTime(new java.util.Date()));
		payment.setOrder(getUser().getOrder());
		payment.setProcessorClassName(this.getClass().getName());
		payment
				.setValueFieldLabels("Order Total,PayPal Token,Payment Status,Message,Payment Type,Transaction Type,Transaction ID,,");
		payment.setValue1((String) parameters.get("orderTotal"));
		payment.setValue2((String) parameters.get("token"));
		payment.setValue3((String) parameters.get("paymentStatus"));
		payment.setValue4("User elected to pay with PayPal");
		payment.setValue5((String) parameters.get("paymentType"));
		payment.setValue6((String) parameters.get("transactionType"));
		payment.setValue7((String) parameters.get("transactionID"));
		PaymentDAO paymentDAO = (PaymentDAO) getDaoFactory().createDAO(
				"paymentDAOImplementer");
		paymentDAO.setPayment(payment);
		paymentDAO.insertPayment(false);
		if (getUser().getOrder().getPayments() == null) {
			getUser().getOrder().setPayments(new LinkedHashSet());
		}
		getUser().getOrder().getPayments().add(payment);
	}

	public String convertStateOrProvice(String rawStateOrProvice) {
		if (rawStateOrProvice == null) {
			return null;
		}
		if (rawStateOrProvice.equals("Alberta")) {
			return "AB";
		}
		if (rawStateOrProvice.equals("British Columbia")) {
			return "BC";
		}
		if (rawStateOrProvice.equals("Manitoba")) {
			return "MB";
		}
		if (rawStateOrProvice.equals("New Brunswick")) {
			return "NB";
		}
		if (rawStateOrProvice.equals("Newfoundland")) {
			return "NF";
		}
		if (rawStateOrProvice.equals("Northwest Territories")) {
			return "NT";
		}
		if (rawStateOrProvice.equals("Nova Scotia")) {
			return "NS";
		}
		if (rawStateOrProvice.equals("Nunavut")) {
			return "NU";
		}
		if (rawStateOrProvice.equals("Ontario")) {
			return "ON";
		}
		if (rawStateOrProvice.equals("Prince Edward Island")) {
			return "PE";
		}
		if (rawStateOrProvice.equals("Quebec")) {
			return "QC";
		}
		if (rawStateOrProvice.equals("Saskatchewan")) {
			return "SK";
		}
		if (rawStateOrProvice.equals("Yukon")) {
			return "YT";
		}
		return rawStateOrProvice;
	}

	private PaymentDetailsItemType[] convertCartToPayPalItems(Order order) {
		PaymentDetailsItemType[] paypalItems = null;
		int numberOfItems = 0;
		if (order != null && order.firstDelivery() != null && order.firstDelivery().getOrderItemCollection() != null) {
			Collection items = order.firstDelivery().getOrderItemCollection();
			numberOfItems = items.size();
		}
		int numberOfGlobalDiscounts = 0;
		if (order != null && order.getOrderDiscounts() != null) {
			Collection discounts = order.getOrderDiscounts();
			Iterator i = discounts.iterator();
			while (i.hasNext()) {
				OrderDiscount od = (OrderDiscount) i.next();
				if (od.getItemsAppliesType().equals("[global]")) {
					numberOfGlobalDiscounts++;
				}
			}
		}
		if (numberOfItems + numberOfGlobalDiscounts > 0) {
			paypalItems = new PaymentDetailsItemType[numberOfItems + numberOfGlobalDiscounts];
		}
		int count = 0;
		if (numberOfItems > 0) {
			Collection items = order.firstDelivery().getOrderItemCollection();
			Iterator i = items.iterator();
			while (i.hasNext()) {
				OrderItem oi = (OrderItem) i.next();
				PaymentDetailsItemType paymentItem = new PaymentDetailsItemType();
				paymentItem.setName(oi.getName());
				paymentItem.setNumber(oi.getCode());
				double itemPrice = utils().roundForMoney(oi.getItemUnitPrice());
				BasicAmountType bat = new BasicAmountType();
				bat.setValue(Double.toString(itemPrice));
				bat.setCurrencyID(CurrencyCodeType.fromValue((String) getSettings().getValue("currencyCode")));
				paymentItem.setAmount(bat);
				paymentItem.setQuantity(oi.getQuantity());
				paypalItems[count] = paymentItem;
				count++;
			}
		}
		if (numberOfGlobalDiscounts > 0) {
			Collection discounts = order.getOrderDiscounts();
			Iterator i = discounts.iterator();
			while (i.hasNext()) {
				OrderDiscount od = (OrderDiscount) i.next();
				if (od.getItemsAppliesType().equals("[global]")) {
					PaymentDetailsItemType paymentItem = new PaymentDetailsItemType();
					paymentItem.setName(od.getName());
					paymentItem.setNumber(od.getCode());
					double itemPrice = utils().roundForMoney(od.getTotal() * -1);
					BasicAmountType bat = new BasicAmountType();
					bat.setValue(Double.toString(itemPrice));
					bat.setCurrencyID(CurrencyCodeType.fromValue((String) getSettings().getValue("currencyCode")));
					paymentItem.setAmount(bat);
					paymentItem.setQuantity(1);
					paypalItems[count] = paymentItem;
					count++;
				}
			}
		}
		return paypalItems;
	}

	private Double determinePayPalOrderTotal(PaymentDetailsItemType[] paypalItems) {
		double itemTotal = 0;
		for (int i = 0; i < paypalItems.length; i++) {
			PaymentDetailsItemType paypalItem = paypalItems[i];
			itemTotal = itemTotal + Double.parseDouble(paypalItem.getAmount().getValue())
					* paypalItem.getQuantity().intValue();
		}
		return new Double(itemTotal);
	}
}
