package com.softslate.commerce.businessobjects.shipping;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.apache.axis.types.PositiveInteger;
import org.apache.axis.utils.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.fedex.addressvalidation.stub.Address;
import com.fedex.addressvalidation.stub.AddressToValidate;
import com.fedex.addressvalidation.stub.AddressValidationAccuracyType;
import com.fedex.addressvalidation.stub.AddressValidationChangeType;
import com.fedex.addressvalidation.stub.AddressValidationOptions;
import com.fedex.addressvalidation.stub.AddressValidationPortType;
import com.fedex.addressvalidation.stub.AddressValidationReply;
import com.fedex.addressvalidation.stub.AddressValidationRequest;
import com.fedex.addressvalidation.stub.AddressValidationResult;
import com.fedex.addressvalidation.stub.AddressValidationServiceLocator;
import com.fedex.addressvalidation.stub.ClientDetail;
import com.fedex.addressvalidation.stub.Notification;
import com.fedex.addressvalidation.stub.NotificationSeverityType;
import com.fedex.addressvalidation.stub.ProposedAddressDetail;
import com.fedex.addressvalidation.stub.TransactionDetail;
import com.fedex.addressvalidation.stub.VersionId;
import com.fedex.addressvalidation.stub.WebAuthenticationCredential;
import com.fedex.addressvalidation.stub.WebAuthenticationDetail;
import com.softslate.commerce.businessobjects.core.BaseBusinessProcessor;

public class FedExAddressValidationProcessor extends BaseBusinessProcessor implements
 AddressValidationProcessor {

	static Log log = LogFactory.getLog(FedExAddressValidationProcessor.class);

	public Map validateAddress(Map parameters) throws Exception {
		if (!areFedExSettingsOk()) {
			log.info("FedEx settings not configured to do address validation - skipping it.");
			return null;
		}
		Map results = new HashMap();
		AddressValidationRequest request = buildAddressValidationRequest(parameters);

		// Initialize the service
		AddressValidationServiceLocator service;
		AddressValidationPortType port;
		service = new AddressValidationServiceLocator();
		updateEndPoint(service);
		port = service.getAddressValidationServicePort();
		AddressValidationReply reply = port.addressValidation(request);

		// Check if the call was successful
		if (isResponseOk(reply.getHighestSeverity())) {
			results.putAll(processAddressValidationResponse(parameters, reply));
		}
		results.putAll(processNotifications(reply.getNotifications()));
		return results;
	}

	protected boolean areFedExSettingsOk() {
		boolean areFedExSettingsOk = false;
		String enableFedExAddressValidation = getSettings().getValue("enableFedExAddressValidation");
		if (enableFedExAddressValidation != null && enableFedExAddressValidation.equals("1")) {
			String accountNumber = getSettings().getValue("fedExAccountNumberAddVal");
			String meterNumber = getSettings().getValue("fedExMeterNumberAddVal");
			if (accountNumber == null || meterNumber == null) {
				accountNumber = getSettings().getValue("fedExAccountNumber");
				meterNumber = getSettings().getValue("fedExMeterNumber");
			}
			String key = getSettings().getValue("fedExAPIKeyAddVal");
			String password = getSettings().getValue("fedExPasswordAddVal");
			if (key == null || password == null) {
				key = getSettings().getValue("fedExAPIKey");
				password = getSettings().getValue("fedExPassword");
			}
			if (!StringUtils.isEmpty(accountNumber) && !StringUtils.isEmpty(meterNumber) && !StringUtils.isEmpty(key)
					&& !StringUtils.isEmpty(password)) {
				areFedExSettingsOk = true;
			}
		}
		return areFedExSettingsOk;
	}

	protected AddressValidationRequest buildAddressValidationRequest(Map parameters) {
		AddressValidationRequest request = new AddressValidationRequest();

		// Set misc properties
		request.setClientDetail(createClientDetail());
		request.setWebAuthenticationDetail(createWebAuthenticationDetail());
		VersionId versionId = new VersionId("aval", 2, 0, 0);
		request.setVersion(versionId);

		// Set transaction detail
		TransactionDetail transactionDetail = new TransactionDetail();
		transactionDetail.setCustomerTransactionId(getUser().getOrder().getOrderNumber() + "-"
				+ System.currentTimeMillis());
		request.setTransactionDetail(transactionDetail);

		// Set timestamp
		Calendar c = Calendar.getInstance();
		request.setRequestTimestamp(c);

		// Set various options
		AddressValidationOptions vo = new AddressValidationOptions();
		vo.setMaximumNumberOfMatches(new PositiveInteger("5"));
		vo.setStreetAccuracy(AddressValidationAccuracyType.MEDIUM);
		vo.setDirectionalAccuracy(AddressValidationAccuracyType.MEDIUM);
		vo.setCompanyNameAccuracy(AddressValidationAccuracyType.MEDIUM);
		// vo.setConvertToUpperCase(new Boolean(true));
		vo.setRecognizeAlternateCityNames(new Boolean(true));
		vo.setReturnParsedElements(new Boolean(true));
		vo.setCheckResidentialStatus(new Boolean(true));
		request.setOptions(vo);

		// Define the customer's address
		AddressToValidate av1 = new AddressToValidate();
		Address customerAddress = new Address();
		if (getUser().getOrder().firstDelivery().getAddress2() != null
				&& getUser().getOrder().firstDelivery().getAddress2().length() > 0) {
			customerAddress.setStreetLines(new String[] { getUser().getOrder().firstDelivery().getAddress1(),
					getUser().getOrder().firstDelivery().getAddress2() });
		} else {
			customerAddress.setStreetLines(new String[] { getUser().getOrder().firstDelivery().getAddress1() });

		}
		customerAddress.setCity(getUser().getOrder().firstDelivery().getCity());
		customerAddress.setStateOrProvinceCode(getUser().getOrder().firstDelivery().getState());
		customerAddress.setPostalCode(getUser().getOrder().firstDelivery().getPostalCode());
		customerAddress.setCountryCode(getUser().getOrder().firstDelivery().getCountry());
		av1.setAddress(customerAddress);
		if (getUser().getOrder().firstDelivery().getOrganization() != null
				&& getUser().getOrder().firstDelivery().getOrganization().length() > 0) {
			av1.setCompanyName(getUser().getOrder().firstDelivery().getOrganization());
		}
		av1.setAddressId("Customer Address");
		request.setAddressesToValidate(new AddressToValidate[] { av1 });
		return request;
	}

	protected ClientDetail createClientDetail() {
		ClientDetail clientDetail = new ClientDetail();
		String accountNumber = (String) getSettings().getValue("fedExAccountNumberAddVal");
		String meterNumber = (String) getSettings().getValue("fedExMeterNumberAddVal");
		if (accountNumber == null || meterNumber == null) {
			accountNumber = (String) getSettings().getValue("fedExAccountNumber");
			meterNumber = (String) getSettings().getValue("fedExMeterNumber");
		}
		clientDetail.setAccountNumber(accountNumber);
		clientDetail.setMeterNumber(meterNumber);
		return clientDetail;
	}

	protected WebAuthenticationDetail createWebAuthenticationDetail() {
		WebAuthenticationCredential wac = new WebAuthenticationCredential();
		String key = (String) getSettings().getValue("fedExAPIKeyAddVal");
		String password = (String) getSettings().getValue("fedExPasswordAddVal");
		if (key == null || password == null) {
			key = (String) getSettings().getValue("fedExAPIKey");
			password = (String) getSettings().getValue("fedExPassword");
		}
		wac.setKey(key);
		wac.setPassword(password);
		return new WebAuthenticationDetail(wac);
	}

	protected void updateEndPoint(AddressValidationServiceLocator serviceLocator) {

		// Always has to be production for address validations
		// String endPoint = (String) getSettings().getValue("fedExURL");
		String endPoint = "https://gateway.fedex.com:443/web-services";
		if (endPoint != null) {
			serviceLocator.setAddressValidationServicePortEndpointAddress(endPoint);
		}
	}

	protected Map processAddressValidationResponse(Map parameters, AddressValidationReply reply) throws Exception {
		Map results = new HashMap();
		StringBuffer logMessage = new StringBuffer("");
		logMessage.append("Address validation details: ");
		AddressValidationResult[] avr = reply.getAddressResults();
		for (int i = 0; i < avr.length; i++) {
			ProposedAddressDetail[] ad1 = avr[i].getProposedAddressDetails();
			for (int j = 0; j < ad1.length; j++) {
				results.put("Score", ad1[j].getScore());
				logMessage.append(" Score - " + ad1[j].getScore());
				try {
					results.put("ResidentialStatus", ad1[j].getResidentialStatus().getValue());
					results.put("Address1", ad1[j].getAddress().getStreetLines(0));
					results.put("City", ad1[j].getAddress().getCity());
					results.put("StateOrProvinceCode", ad1[j].getAddress().getStateOrProvinceCode());
					results.put("PostalCode", ad1[j].getAddress().getPostalCode());
					results.put("CountryCode", ad1[j].getAddress().getCountryCode());
					logMessage.append("; Address - " + ad1[j].getAddress().getStreetLines(0));
					logMessage.append("; StateOrProvinceCode - " + ad1[j].getAddress().getStateOrProvinceCode());
					logMessage.append("; City - " + ad1[j].getAddress().getCity());
					logMessage.append("; PostalCode - " + ad1[j].getAddress().getPostalCode());
					logMessage.append("; CountryCode - " + ad1[j].getAddress().getCountryCode());
					logMessage.append("; Changes - " + ad1[j].getScore());
					AddressValidationChangeType[] avct = ad1[j].getChanges();
					for (int k = 0; k < avct.length; k++) {
						results.put(avct[k].toString(), Boolean.TRUE);
						logMessage.append("; " + avct[k].toString());
					}
				} catch (Exception e) {
					log.warn("Exception parsing FedEx address validation response: ", e);
					log.info(logMessage.toString());
				}
			}
		}
		if (log.isInfoEnabled()) {
			log.info(logMessage.toString());
		}
		return results;
	}

	protected boolean isResponseOk(NotificationSeverityType notificationSeverityType) {
		if (notificationSeverityType == null) {
			return false;
		}
		if (notificationSeverityType.equals(NotificationSeverityType.WARNING)
				|| notificationSeverityType.equals(NotificationSeverityType.NOTE)
				|| notificationSeverityType.equals(NotificationSeverityType.SUCCESS)) {
			return true;
		}
		return false;
	}

	protected Map processNotifications(Notification[] notifications) {
		Map results = new HashMap();
		Collection messages = new ArrayList();
		Collection errors = new ArrayList();
		if (notifications == null || notifications.length == 0) {
			log.info("No notifications returned");
		}
		for (int i = 0; i < notifications.length; i++) {
			Notification n = notifications[i];
			NotificationSeverityType nst = n.getSeverity();

			StringBuffer sb = new StringBuffer("");
			sb.append(n.getMessage());
			sb.append(" (Severity: " + (nst == null ? "null" : nst.getValue()));
			sb.append(" Code: " + n.getCode());
			sb.append(" Source: " + n.getSource() + ")");
			if (nst.getValue() != null && nst.equals(NotificationSeverityType.ERROR)) {
				log.error(sb.toString());
				if (n.getCode().equals("8336")) {
					errors.add("Delivery method not allowed to the specified address. Please choose another.");
				} else {
					errors.add(n.getMessage());
				}
			} else if (nst.getValue() != null && nst.equals(NotificationSeverityType.FAILURE)) {
				log.error(sb.toString());
				errors.add(n.getMessage());
			} else if (nst.getValue() != null && nst.equals(NotificationSeverityType.WARNING)) {
				log.warn(sb.toString());
				messages.add(n.getMessage());
			} else if (nst.getValue() != null && !nst.equals(NotificationSeverityType.SUCCESS)) {
				log.info(sb.toString());
				messages.add(n.getMessage());
			} else if (nst.getValue() != null && nst.equals(NotificationSeverityType.SUCCESS)) {
				log.info(sb.toString());
			}
		}
		results.put("errors", errors);
		results.put("messages", messages);
		return results;
	}
}
