package com.softslate.commerce.customer.product;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;

import com.softslate.commerce.businessobjects.product.LuceneSearcher;
import com.softslate.commerce.businessobjects.product.Manufacturer;
import com.softslate.commerce.businessobjects.product.ManufacturerProcessor;
import com.softslate.commerce.businessobjects.product.ProductProcessor;
import com.softslate.commerce.businessobjects.product.SearchList;
import com.softslate.commerce.businessobjects.product.SearchListBean;
import com.softslate.commerce.customer.core.BaseAction;
import com.softslate.commerce.customer.core.BaseDynaForm;

/**
 * Struts action class that corresponds to the manufacturer page. Used for
 * requests to "/Manufacturer.do".
 * <p>
 * The <code>execute</code> method of this class populates a manufacturer's
 * information from the database and puts it in the request and session scopes
 * for use by the view layer. Given a
 * {@link com.softslate.commerce.customer.product.ManufacturerForm} with its
 * <code>code</code> property populated, it creates and initializes an instance
 * of {@link com.softslate.commerce.businessobjects.product.Manufacturer}. It
 * then invokes
 * {@link com.softslate.commerce.businessobjects.product.ManufacturerProcessor#getManufacturerFromCode(Manufacturer)}
 * to load the manufacturer's information from the database.
 * <p>
 * In the default implementation, in addition to all of the manufacturer's
 * regular properties, its <code>productListCollection</code> is populated with
 * just those <code>Manufacturer</code>s falling on the current page being
 * displayed. The paging parameter for the number of items to display comes from
 * the "manufacturerPageSize" setting in the <code>sscSetting</code> database
 * table. Which page to display is determined by the value of the
 * <code>firstRow</code> request parameter (which is held by the incoming
 * <code>ManufacturerForm</code> and copied to the <code>Manufacturer</code>
 * object being populated).
 * <p>
 * This method puts the initialized <code>Manufacturer</code> object in the
 * request scope as an attribute with the name "manufacturer".
 * <p>
 * The <code>execute</code> method next loads a <code>Map</code> of
 * attribute-only SKUs and places it in the request as an attribute named
 * "attributeSKUs" so they may be displayed on the product page. (Attribute-only
 * SKUs are SKUs which are not tied to any particular product but which may
 * still be relevant to the one of the category's products.)
 * 
 * @author David Tobey
 * 
 * @see com.softslate.commerce.customer.product.ManufacturerForm
 * @see com.softslate.commerce.businessobjects.product.Manufacturer
 * @see com.softslate.commerce.businessobjects.product.ManufacturerProcessor
 */
public class ManufacturerAction extends BaseAction {

	static Log log = LogFactory.getLog(ManufacturerAction.class);

	public ActionForward execute(ActionMapping mapping, ActionForm form, HttpServletRequest request,
			HttpServletResponse response) throws Exception {
		if (log.isInfoEnabled())
			log.info("Starting action.");
		BaseDynaForm baseForm = (BaseDynaForm) form;
		ManufacturerProcessor manufacturerProcessor =  baseForm.getBusinessObjectFactory()
				.createObject(ManufacturerProcessor.class);
		ProductProcessor productProcessor = (ProductProcessor) baseForm.getBusinessObjectFactory().createObject(
				"productProcessorImplementer");

		// Initialize a Manufacturer object
		Manufacturer manufacturer = initializeManufacturer(baseForm);
		boolean shouldProductsBeLoadedFromDB = shouldProductsBeLoadedFromDB(baseForm);
		manufacturer = manufacturerProcessor.getManufacturerFromCode(manufacturer, shouldProductsBeLoadedFromDB);
		if (manufacturer.getManufacturerID() == 0) {
			baseForm.getErrors().add(ActionMessages.GLOBAL_MESSAGE,
					new ActionMessage("errors.manufacturerNotFound", manufacturer.getCode()));
			saveErrors(baseForm.getRequest(), (ActionMessages) baseForm.getErrors());
			response.setStatus(HttpServletResponse.SC_NOT_FOUND);
			return mapping.findForward("failure");
		}

		// Load the products via Lucene if they weren't loaded from the DB
		SearchList searchList = null;
		if (!shouldProductsBeLoadedFromDB) {
			searchList = loadProductsWithAdvancedSearch(baseForm, manufacturer, (String) request.getParameter("sortBy"));
		}

		// Store objects in the request scope for use by JSPs
		loadAttributeOnlySKUs(baseForm, request, productProcessor);
		saveObjectsInRequestScope(manufacturer, searchList, request, baseForm);
		defineMetaTags(manufacturer, request);

		// Forward to Ajax tile if it's an Ajax request (faceted search)
		if (baseForm.getMap().get("ajax") != null && baseForm.getMap().get("ajax").equals("true")) {
			if (("default-html5").equals(baseForm.html5OrXhtml())) {
				return mapping.findForward("successAjaxHTML5");
			} else {
				return mapping.findForward("successAjax");
			}
		}

		return mapping.findForward("success");
	}

	protected Manufacturer initializeManufacturer(BaseDynaForm baseForm) throws Exception {
		Manufacturer manufacturer = (Manufacturer) baseForm.getBusinessObjectFactory().createObject(
				"manufacturerImplementer");
		BeanUtils.copyProperties(manufacturer, baseForm);

		// Set the product list parameters
		String itemsPerPage = (String) baseForm.getSettings().getValue("manufacturerPageSize");

		// 'View all' support
		if (baseForm.getMap().get("firstRow") != null && baseForm.getMap().get("firstRow").equals("all")) {
			itemsPerPage = (String) baseForm.getSettings().getValue("maxRecordsProcessed");
			baseForm.getMap().put("firstRow", "1");
		}
		if (log.isDebugEnabled())
			log.debug("itemsPerPage: " + itemsPerPage);
		manufacturer.setItemsPerPage(Integer.parseInt(itemsPerPage));
		return manufacturer;
	}

	protected boolean shouldProductsBeLoadedFromDB(BaseDynaForm baseForm) {
		String advancedSearchManufacturer = (String) baseForm.getSettings().getValue("advancedSearchManufacturer");
		boolean forceLoadProducts = baseForm.getMap().get("forceLoadProducts") != null
				&& baseForm.getMap().get("forceLoadProducts").equals("true");
		boolean shouldProductsBeLoadedFromDB = advancedSearchManufacturer.equals("0") || forceLoadProducts;
		return shouldProductsBeLoadedFromDB;
	}

	protected SearchList loadProductsWithAdvancedSearch(BaseDynaForm baseForm, Manufacturer manufacturer,
			String sortByParameter) throws Exception {
		// First create the processor to do the searching
		LuceneSearcher luceneSearcher = (LuceneSearcher) baseForm.getBusinessObjectFactory().createObject(
				"luceneSearcherImplementer");

		// Create a SearchList helper bean
		SearchList searchList =  baseForm.getBusinessObjectFactory().createObject(SearchList.class);
		BeanUtils.copyProperties(searchList, baseForm);
		baseForm.getFormUtils().eliminateDuplicateFacets(searchList);

		// Set the path to the index
		String index = (String) baseForm.getSettings().getValue("luceneIndexDir");
		searchList.setIndexDir(getServlet().getServletContext().getRealPath(index));
		if (baseForm.getMap().get("firstRow") != null && baseForm.getMap().get("firstRow").equals("all")) {
			searchList.setFirstRow(1);
		}
		searchList.setItemsPerPage(manufacturer.getItemsPerPage());
		searchList.setSearchListType(SearchListBean.SEARCH_PRODUCTS_WITHIN_MANUFACTURER);
		searchList.setManufacturerID(manufacturer.getManufacturerID());
		if (sortByParameter != null) {
			searchList.setSortBy(sortByParameter);
		} else {
			searchList.setSortBy("productOrder");
		}
		searchList.setCategoryMap((Map) getServlet().getServletContext().getAttribute("categoryMap"));

		// Load the objects and place the list in the request scope
		searchList = luceneSearcher.getSearchList(searchList);

		return searchList;
	}

	protected void loadAttributeOnlySKUs(BaseDynaForm baseForm, HttpServletRequest request,
			ProductProcessor productProcessor) throws Exception {
		String trackInventory = (String) baseForm.getSettings().getValue("trackInventory");
		String applyDiscounts = (String) baseForm.getSettings().getValue("applyDiscounts");
		if ((trackInventory != null && trackInventory.equals("1"))
				|| (applyDiscounts != null && applyDiscounts.equals("1"))) {
			Map attributeSKUs = productProcessor.loadAttributeSKUs();
			if (log.isDebugEnabled())
				log.debug("Atribute-only skus: " + attributeSKUs);
			request.setAttribute("attributeSKUs", attributeSKUs);
			Collection attributeSKUCollection = new ArrayList();
			if (!attributeSKUs.isEmpty()) {
				attributeSKUCollection = productProcessor.loadAttributeSKUCollection();
			}
			request.setAttribute("attributeSKUCollection", attributeSKUCollection);
		}
	}

	protected void saveObjectsInRequestScope(Manufacturer manufacturer, SearchList searchList,
			HttpServletRequest request, BaseDynaForm baseForm) throws Exception {
		request.setAttribute("manufacturer", manufacturer);
		request.setAttribute("manufacturerParam", "&manufacturer=" + manufacturer.getCode());
		request.setAttribute("linkBase", "/Manufacturer.do?code=" + manufacturer.getCode() + "&");
		if (searchList != null) {
			request.setAttribute("searchList", searchList);
			request.setAttribute("advancedSearch", "true");
			request.setAttribute("itemCount", new Integer(searchList.getItemCount()));
			request.setAttribute("itemsPerPage", new Integer(searchList.getItemsPerPage()));
			request.setAttribute("firstRow", new Integer(searchList.getFirstRow()));
			request.setAttribute("listCollection", searchList.getSearchListCollection());
			request.setAttribute("objectTypes", searchList.getObjectTypes());
			String sort = searchList.getSortBy();
			request.setAttribute("sortByParam", "&sortBy=" + java.net.URLEncoder.encode(sort, "UTF-8"));
			String facetsParam = baseForm.getFormUtils().stringArrayAsURLParam("facets", searchList.getFacets());
			request.setAttribute("facetsParam", facetsParam);
		} else {
			request.setAttribute("itemCount", new Integer(manufacturer.getItemCount()));
			request.setAttribute("itemsPerPage", new Integer(manufacturer.getItemsPerPage()));
			request.setAttribute("firstRow", new Integer(manufacturer.getFirstRow()));
			request.setAttribute("listCollection", manufacturer.getProductListCollection());
		}
	}

	protected void defineMetaTags(Manufacturer manufacturer, HttpServletRequest request) {

		// Set various meta tag fields
		if (manufacturer.getName() != null) {
			String subTitle = manufacturer.getName().replaceAll("\\<.*?>", "");
			if (subTitle.length() > 0) {
				request.setAttribute("htmlHeadTitle", request.getAttribute("htmlHeadTitle") + " - " + subTitle);
			}
		}
		if (manufacturer.getDescription() != null) {
			String desc = manufacturer.getDescription().replaceAll("\\<.*?>", "");
			if (desc.length() > 0) {
				request.setAttribute("htmlHeadDescription", desc);
			}
		}
	}

}
